from typing import IO, cast
from smart_open import open as _smart_open


def smart_open(
    uri,
    mode="r",
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    closefd=True,
    opener=None,
    ignore_ext=False,
    transport_params=None,
) -> IO[str]:
    return cast(
        IO,
        _smart_open(
            uri,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
            ignore_ext=ignore_ext,
            transport_params=transport_params,
        ),
    )



from json import JSONEncoder

from web3.types import HexBytes
from web3.datastructures import AttributeDict


class EthJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, HexBytes):
            return o.hex()
        if isinstance(o, AttributeDict):
            return dict(o)
        if isinstance(o, bytes):
            return o.hex()
        return super().default(o)



import os
from argparse import ArgumentParser

from eth_tools import commands
from eth_tools import constants
from eth_tools.contract_caller import DEFAULT_BLOCK_INTERVAL


def environ_or_required(key):
    default_value = os.environ.get(key)
    if default_value:
        return {"default": default_value}
    return {"required": True}


def add_web3_uri(subparser):
    subparser.add_argument(
        "--web3-uri", help="URI of Web3", **environ_or_required("WEB3_PROVIDER_URI")
    )


def add_etherscan_api_key(subparser):
    subparser.add_argument(
        "--etherscan-api-key",
        help="API key for Etherscan",
        **environ_or_required("ETHERSCAN_API_KEY"),
    )


parser = ArgumentParser(prog="ethereum-tools")

subparsers = parser.add_subparsers(dest="command", help="Command to execute")

fetch_block_timestamps_parser = subparsers.add_parser(
    "fetch-blocks", help="fetches information about given blocks"
)
add_web3_uri(fetch_block_timestamps_parser)
fetch_block_timestamps_parser.add_argument(
    "-f",
    "--fields",
    nargs="+",
    default=constants.DEFAULT_BLOCK_FIELDS,
    help="fields to fetch from the block",
)
fetch_block_timestamps_parser.add_argument(
    "-s",
    "--start-block",
    type=int,
    default=1,
    help="block from which to fetch timestamps",
)
fetch_block_timestamps_parser.add_argument(
    "-e", "--end-block", type=int, help="block up to which to fetch timestamps"
)
fetch_block_timestamps_parser.add_argument(
    "-o", "--output", required=True, help="output file"
)
fetch_block_timestamps_parser.add_argument(
    "--log-interval", type=int, default=1_000, help="interval at which to log"
)

fetch_address_transactions_parser = subparsers.add_parser(
    "fetch-address-transactions",
    help="fetches information about the transactions of a given address from Etherscan",
)
add_etherscan_api_key(fetch_address_transactions_parser)
fetch_address_transactions_parser.add_argument(
    "-a", "--address", required=True, help="address for which to get transactions"
)
fetch_address_transactions_parser.add_argument(
    "--internal",
    default=False,
    action="store_true",
    help="fetch information about internal transactions rather than regular ones",
)
fetch_address_transactions_parser.add_argument(
    "-o", "--output", required=True, help="output file"
)

fetch_transactions_parser = subparsers.add_parser(
    "fetch-transactions",
    help="fetches transactions in given files from an Ethereum node",
)
add_web3_uri(fetch_transactions_parser)
fetch_transactions_parser.add_argument(
    "files", nargs="+", help="files containing transactions to trace"
)
fetch_transactions_parser.add_argument(
    "-r",
    "--include-receipt",
    action="store_true",
    default=False,
    help="include transaction receipt",
)
fetch_transactions_parser.add_argument(
    "-t",
    "--include-traces",
    action="store_true",
    default=False,
    help="include transaction traces",
)
fetch_transactions_parser.add_argument(
    "-o", "--output", required=True, help="output file"
)

call_contract_parser = subparsers.add_parser(
    "call-contract", help="call contract regularly between blocks"
)
add_web3_uri(call_contract_parser)
call_contract_parser.add_argument("address", help="address of the contract")
call_contract_parser.add_argument("--abi", help="path to the contract abi")
call_contract_parser.add_argument("-s", "--start", type=int, help="start block")
call_contract_parser.add_argument("-e", "--end", type=int, help="end block")
call_contract_parser.add_argument(
    "-i",
    "--interval",
    type=int,
    default=DEFAULT_BLOCK_INTERVAL,
    help="interval between calls",
)
call_contract_parser.add_argument(
    "-f", "--func", required=True, help="function to call"
)
call_contract_parser.add_argument(
    "--args", nargs="*", help="arguments to pass to the function"
)
call_contract_parser.add_argument("-o", "--output", help="output file")


fetch_events_parser = subparsers.add_parser(
    "fetch-events", help="Fetches the events from the given contract"
)
add_web3_uri(fetch_events_parser)
fetch_events_parser.add_argument("address", help="Address of the contract")
fetch_events_parser.add_argument("--abi", help="Path to the contract ABI")
fetch_events_parser.add_argument(
    "-s",
    "--start-block",
    help="Start block to fetch the events",
    required=True,
    type=int,
)
fetch_events_parser.add_argument(
    "-e", "--end-block", help="End block to fetch the events", type=int
)
fetch_events_parser.add_argument(
    "-o",
    "--output",
    required=True,
    help="Output jsonl file to store the results (.gz recommended)",
)

bulk_fetch_events_parser = subparsers.add_parser(
    "bulk-fetch-events", help="Fetches the events from the given contracts"
)
add_web3_uri(bulk_fetch_events_parser)
bulk_fetch_events_parser.add_argument(
    "-c", "--config", help="Config file to fetch events"
)
bulk_fetch_events_parser.add_argument("--abis", help="Directory containing ABIs")
bulk_fetch_events_parser.add_argument(
    "-o",
    "--output",
    required=True,
    help="Output directory to store the results",
)


def run():
    args = vars(parser.parse_args())
    if not args["command"]:
        parser.error("no command given")

    func = getattr(commands, args["command"].replace("-", "_"))
    func(args)



from web3 import Web3
from retry import retry

from eth_tools.logger import logger


class TransactionTracer:
    def __init__(self, web3: Web3):
        self.web3 = web3

    @retry(delay=1, backoff=2, tries=3, logger=logger)
    def trace_transaction(self, tx_hash: str, disable_memory=True, disable_storage=True):
        params = [tx_hash, {"disableMemory": disable_memory, "disableStorage": disable_storage}]
        return self.web3.manager.request_blocking("debug_traceTransaction", params=params)



import csv
import json
import sys
from contextlib import contextmanager
from functools import wraps
from typing import IO, Iterator

from eth_typing import Address
from web3 import Web3
from web3.providers.auto import load_provider_from_uri

from eth_tools import constants
from eth_tools.block_iterator import BlockIterator
from eth_tools.contract_caller import ContractCaller
from eth_tools.event_fetcher import EventFetcher, FetchTask
from eth_tools.json_encoder import EthJSONEncoder
from eth_tools.logger import logger
from eth_tools.transaction_fetcher import TransactionsFetcher
from eth_tools.transaction_tracer import TransactionTracer
from eth_tools.utils import smart_open


def uses_web3(f):
    @wraps(f)
    def wrapper(args):
        web3 = create_web3(args["web3_uri"])
        return f(args, web3)

    return wrapper


def uses_etherscan(f):
    @wraps(f)
    def wrapper(args):
        etherscan_key = args["etherscan_api_key"]
        return f(args, etherscan_key)

    return wrapper


def create_web3(uri: str):
    provider = load_provider_from_uri(uri, {"timeout": 60})
    return Web3(provider=provider)


@uses_web3
def fetch_blocks(args: dict, web3: Web3):
    """Fetches blocks and stores them in the file given in arguments"""
    block_iterator = BlockIterator(
        web3, args["start_block"], args["end_block"], log_interval=args["log_interval"]
    )
    fields = args["fields"]
    with smart_open(args["output"], "w") as f:
        writer = csv.DictWriter(f, fieldnames=fields)
        writer.writeheader()
        for block in block_iterator:
            row = {field: getattr(block, field) for field in fields}
            writer.writerow(row)


@uses_etherscan
def fetch_address_transactions(args: dict, etherscan_key: Web3):
    fetcher = TransactionsFetcher(etherscan_api_key=etherscan_key)
    internal = args["internal"]
    if internal:
        fields = constants.ETHERSCAN_INTERNAL_TRANSACTION_KEYS
    else:
        fields = constants.ETHERSCAN_TRANSACTION_KEYS
    with smart_open(args["output"], "w") as f:
        writer = csv.DictWriter(f, fieldnames=fields)
        writer.writeheader()
        for transaction in fetcher.fetch_contract_transactions(
            args["address"], internal=internal
        ):
            writer.writerow(transaction)


@uses_web3
def fetch_transactions(args: dict, web3: Web3):
    tx_hashes = []
    for filename in args["files"]:
        with smart_open(filename) as fin:
            for tx in csv.DictReader(fin):
                tx_hashes.append(tx["hash"])

    tx_tracer = TransactionTracer(web3)
    done = set()
    with smart_open(args["output"], "w") as fout:
        for i, tx_hash in enumerate(tx_hashes):
            if i % 10 == 0:
                logger.info("progress: %s/%s", i, len(tx_hashes))
            if tx_hash in done:
                continue
            try:
                tx = dict(web3.eth.getTransaction(tx_hash))
                if args["include_receipt"]:
                    tx["receipt"] = web3.eth.getTransactionReceipt(tx_hash)
                if args["include_traces"]:
                    tx["traces"] = tx_tracer.trace_transaction(tx_hash)
                json.dump(tx, fout, cls=EthJSONEncoder)
                print(file=fout)
                done.add(tx_hash)
            except Exception as ex:  # pylint: disable=broad-except
                logger.warning("failed to trace %s: %s", tx_hash, ex)
                continue


@uses_web3
def call_contract(args: dict, web3: Web3):
    with open(args["abi"]) as f:
        abi = json.load(f)
    address: Address = web3.toChecksumAddress(args["address"])
    contract = web3.eth.contract(abi=abi, address=address)
    contract_caller = ContractCaller(contract)
    with smart_open_with_stdout(args["output"], "w") as fout:
        # print(args["args"])
        results = contract_caller.collect_results(
            args["func"],
            start_block=args["start"],
            end_block=args["end"],
            block_interval=args["interval"],
            contract_args=args["args"],
        )
        for block, result in results:
            line = {"block": block, "result": result}
            json.dump(line, fout, cls=EthJSONEncoder)
            print(file=fout)


@uses_web3
def fetch_events(args: dict, web3: Web3):
    fetcher = EventFetcher(web3)
    task = FetchTask.from_dict(args)
    fetcher.fetch_and_persist_events(task, args["output"])


@uses_web3
def bulk_fetch_events(args: dict, web3: Web3):
    fetcher = EventFetcher(web3)
    with smart_open(args["config"]) as f:
        raw_tasks = json.load(f)
    tasks = [FetchTask.from_dict(raw_task, args["abis"]) for raw_task in raw_tasks]
    fetcher.fetch_all_events(tasks, args["output"])


@contextmanager
def smart_open_with_stdout(filename, mode="r", **kwargs) -> Iterator[IO]:
    if filename is None:
        yield sys.stdout
    else:
        with smart_open(filename, mode, **kwargs) as f:
            yield f



import logging

from eth_tools import constants


def _create_logger():
    logger = logging.Logger("eth-tools") # pylint: disable=redefined-outer-name

    formatter = logging.Formatter(constants.LOG_FORMAT)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)

    return logger


logger = _create_logger()



from concurrent.futures import ThreadPoolExecutor
import multiprocessing

from web3 import Web3
from web3.contract import Contract
from retry import retry

from eth_tools.logger import logger

DEFAULT_BLOCK_INTERVAL = 1_000


ARG_TYPES = {
    "u256": int,
    "address": Web3.toChecksumAddress,
}


class ContractCaller:
    def __init__(self, contract: Contract):
        self.contract = contract

    def collect_results(self, func_name, start_block,
                        end_block=None, block_interval=DEFAULT_BLOCK_INTERVAL,
                        contract_args=None):
        max_workers = multiprocessing.cpu_count() * 5
        if end_block is None:
            end_block = self.contract.web3.eth.blockNumber
        if start_block is None:
            start_block = end_block
        if contract_args is None:
            contract_args = []
        contract_args = [self.transform_arg(arg) for arg in contract_args]

        def run_task(block):
            try:
                return self.call_func(func_name, block, contract_args)
            except Exception as ex: # pylint: disable=broad-except
                logger.error("failed to fetch block %s: %s", block, ex)

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            blocks = range(start_block, end_block + 1, block_interval)
            total_count = len(blocks)
            results = executor.map(run_task, blocks)
            for i, (block, result) in enumerate(zip(blocks, results)):
                if i % 10 == 0 and total_count > 10:
                    logger.info("progress: %s/%s (%.2f%%)", i, total_count, i / total_count * 100)
                if result is not None:
                    yield (block, result)

    @retry(delay=1, backoff=2, tries=3, logger=logger)
    def call_func(self, func_name, block, contract_args):
        func = getattr(self.contract.functions, func_name)
        return func(*contract_args).call(block_identifier=block)

    def transform_arg(self, raw_arg: str):
        args = raw_arg.split(":")
        if len(args) == 1:
            return args[0]
        arg, arg_type = args
        cast = ARG_TYPES.get(arg_type)
        if not cast:
            raise ValueError(f"unknown type {arg_type}")
        return cast(arg)



from unittest.mock import MagicMock, call

import pytest

from eth_tools import transaction_fetcher as tx_fetcher

DUMMY_ADDRESS = "0x" + ("0" * 40)
MAX_CALLS = tx_fetcher.MAX_TRANSACTIONS // tx_fetcher.TRANSACTIONS_PER_PAGE + 1


@pytest.fixture
def fetcher():
    return tx_fetcher.TransactionsFetcher(etherscan_api_key="dummy-api-key")


def make_call(page, internal=False):
    return call(DUMMY_ADDRESS, page, internal=internal)

def make_calls(start_page, end_page, internal=False):
    return [call(DUMMY_ADDRESS, page, internal=internal)
            for page in range(start_page, end_page + 1)]


@pytest.mark.parametrize("responses,calls,txs_count", [
    ([["abc", "def", "ghi"], []], make_calls(1, 2), 3),
    ([["abc", "def", "ghi"], ["jkl"], []], make_calls(1, 3), 4),
    ([["abc"]] * (MAX_CALLS + 1), make_calls(1, 10), 10),
])
def test_fetch_contract_transactions(fetcher, responses, calls, txs_count):
    fetcher._make_transactions_request = MagicMock()
    fetcher._make_transactions_request.side_effect = responses
    result = list(fetcher.fetch_contract_transactions(DUMMY_ADDRESS))
    fetcher._make_transactions_request.assert_has_calls(calls)
    assert len(result) == txs_count



import os

import requests

from eth_tools.logger import logger


ETHERSCAN_API_URL = "https://api.etherscan.io/api"
TRANSACTIONS_PER_PAGE = 1_000
MAX_TRANSACTIONS = 10_000


class TransactionsFetcher:
    """Retrieves transactions from Etherescan
    """
    def __init__(self, etherscan_api_key=None):
        if not etherscan_api_key:
            etherscan_api_key = os.environ.get("ETHERSCAN_API_KEY")
        if not etherscan_api_key:
            raise ValueError("ETHERSCAN_API_KEY not provided")
        self.etherscan_api_key = etherscan_api_key
        self.session = requests.Session()

    def _make_transactions_request(self, address, page, internal=False):
        action = "txlistinternal" if internal else "txlist"
        response = self.session.get(ETHERSCAN_API_URL, params=dict(
            module="account",
            action=action,
            address=address,
            offset=TRANSACTIONS_PER_PAGE,
            page=page,
            sort="desc",
            apikey=self.etherscan_api_key))
        if response.status_code != 200:
            raise ValueError("failed to execute request: {0}".format(response.status_code))
        return response.json()["result"]

    def fetch_contract_transactions(self, address, internal=False):
        logger.debug("getting transactions (internal=%s) for %s", internal, address)
        count = 0
        for page in range(1, MAX_TRANSACTIONS // TRANSACTIONS_PER_PAGE + 1):
            logger.debug("requesting page %d for %s", page, address)
            returned_results = self._make_transactions_request(address, page, internal=internal)
            if not returned_results:
                break
            for transaction in returned_results:
                yield transaction
            count += len(returned_results)
        else:
            logger.warning("more than %s transactions for %s, fetched first %s transactions",
                           MAX_TRANSACTIONS, address, count)
        logger.info("fetched %s transactions (internal=%s) for %s",
                    count, internal, address)



from argparse import RawTextHelpFormatter
import json
import math
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from os import path
from typing import Iterable, Iterator, List, Optional

from eth_typing import Address
from web3 import Web3
from web3.contract import Contract
from web3.types import FilterParams, LogReceipt

from eth_tools.json_encoder import EthJSONEncoder
from eth_tools.logger import logger
from eth_tools.utils import smart_open


@dataclass(repr=False)
class FetchTask:
    address: str
    abi: dict
    start_block: int
    end_block: Optional[int] = None
    name: Optional[str] = None

    @property
    def checksum_address(self) -> Address:
        return Web3.toChecksumAddress(self.address)

    def __repr__(self) -> str:
        return (
            f"FetchTask(address='{self.address}', "
            f"start_block={self.start_block}, end_block={self.end_block})"
        )

    @property
    def display_name(self):
        if self.name:
            return self.name
        return self.address

    @classmethod
    def from_dict(cls, raw_task: dict, abi_paths: str = None):
        raw_task = raw_task.copy()
        default_abi_name = raw_task.get("name", raw_task["address"]).lower()
        abi_path = raw_task.pop("abi", default_abi_name + ".json")
        if abi_paths:
            abi_path = path.join(abi_paths, abi_path)
        with smart_open(abi_path) as f:
            raw_task["abi"] = json.load(f)
        keys = ["address", "abi", "start_block", "end_block", "name"]
        return cls(**{k: raw_task.get(k) for k in keys})


class ContractFetcher:
    BLOCK_GRANULARITIES = [10_000, 1_000, 100, 10, 1]

    def __init__(self, contract: Contract):
        self.contract = contract
        contract_events = [event() for event in self.contract.events]  # type: ignore
        self.events_by_topic = {
            event.build_filter().topics[0]: event
            for event in contract_events
            if not event.abi["anonymous"]
        }


    def process_logs(self, events: List[LogReceipt]) -> List[LogReceipt]:
        results = []
        for event in events:
            topics = event.get("topics")
            if topics and topics[0].hex() in self.events_by_topic:
                event = self.events_by_topic[topics[0].hex()].processLog(event)
            results.append(event)
        return results

    def _fetch_events(self, start_block: int, end_block: int) -> Iterable[LogReceipt]:
        event_filter = self.contract.web3.eth.filter(
            FilterParams(
                address=self.contract.address, fromBlock=start_block, toBlock=end_block
            )
        )
        events = event_filter.get_all_entries()
        return self.process_logs(events)

    def _fetch_batch_parallel(
        self, start_blocks: List[int], end_blocks: List[int]
    ) -> Iterable[LogReceipt]:
        if len(start_blocks) == 1:
            return self._fetch_events(start_blocks[0], end_blocks[0])
        result = []
        with ThreadPoolExecutor(max_workers=20) as executor:
            for events in executor.map(self._fetch_events, start_blocks, end_blocks):
                result.extend(events)
        return result

    def _fetch_batch(self, start_block: int, end_block: int) -> Iterable[LogReceipt]:
        for granularity in self.BLOCK_GRANULARITIES:
            block_count = end_block - start_block + 1
            iterations = range(math.ceil(block_count / granularity))
            start_blocks = [(start_block + i * granularity) for i in iterations]
            end_blocks = [min(b + granularity - 1, end_block) for b in start_blocks]
            try:
                return self._fetch_batch_parallel(start_blocks, end_blocks)
            except Exception:  # pylint: disable=broad-except
                pass
        raise ValueError(
            f"unable to fetch events for {self.contract.address} "
            f"from block {start_block} to {end_block}"
        )

    def fetch_events(
        self, start_block: int, end_block: int = None
    ) -> Iterator[LogReceipt]:
        if end_block is None:
            end_block = self.contract.web3.eth.blockNumber

        block_count = end_block - start_block + 1
        granularity = self.BLOCK_GRANULARITIES[0]

        for i in range(math.ceil(block_count / granularity)):
            logger.info(
                "%s progress: %s/%s",
                self.contract.address,
                i * granularity,
                block_count,
            )
            batch_start_block = start_block + i * granularity
            batch_end_block = min(batch_start_block + granularity - 1, end_block)
            yield from self._fetch_batch(batch_start_block, batch_end_block)


class EventFetcher:
    def __init__(self, web3: Web3):
        self.web3 = web3

    def fetch_events(self, task: FetchTask) -> Iterator[LogReceipt]:
        contract = self.web3.eth.contract(address=task.checksum_address, abi=task.abi)
        fetcher = ContractFetcher(contract)
        return fetcher.fetch_events(task.start_block, task.end_block)

    def fetch_and_persist_events(self, task: FetchTask, output_file: str):
        with smart_open(output_file, "w") as f:
            for event in self.fetch_events(task):
                print(json.dumps(event, cls=EthJSONEncoder), file=f)

    def fetch_all_events(self, fetch_tasks: List[FetchTask], output_directory: str):
        with ThreadPoolExecutor() as executor:
            filepaths = [
                path.join(output_directory, task.display_name) + ".jsonl.gz"
                for task in fetch_tasks
            ]
            futures = {
                executor.submit(self.fetch_and_persist_events, task, output): task
                for task, output in zip(fetch_tasks, filepaths)
            }
            for future in as_completed(futures):
                task = futures[future]
                ex = future.exception()
                if ex:
                    logger.error(
                        "failed to process %s (%s): %s", task.name, task.address, ex
                    )
                else:
                    logger.info("completed to process %s (%s)", task.name, task.address)



from web3 import Web3
from web3.types import BlockData

from eth_tools.logger import logger


class Block:
    """Wrapper of web3.types.BlockData prodiving a ``transactions_count`` property
    """
    def __init__(self, data: BlockData):
        self.data = data

    def __getattr__(self, key):
        if key not in self.data:
            raise AttributeError
        value = self.data[key]
        if hasattr(value, "hex"):
            value = value.hex()
        return value

    @property
    def transactions_count(self):
        return len(self.transactions)

    def __repr__(self):
        return "Block(number={number})".format(**self.data)


class BlockIterator:
    """Iterator over Ethereum blocks
    It will lazily fetch each block from ``start_block`` to ``end_block`` inclusive
    using the provided ``web3`` instance.
    """
    def __init__(self, web3: Web3, start_block: int = 0, end_block: int = None,
                 log_interval: int = None):
        self.web3 = web3
        self.start_block = start_block
        if end_block is None:
            end_block = self.web3.eth.blockNumber
        self.end_block = end_block
        self.current_block = self.start_block
        self.log_interval = log_interval

    def __iter__(self):
        logger.info("processing %s blocks", self.blocks_count)
        return self

    @property
    def blocks_count(self):
        return self.end_block - self.start_block + 1

    @property
    def processed_count(self):
        return self.current_block - self.start_block

    def __next__(self) -> Block:
        if self.current_block > self.end_block:
            raise StopIteration
        if self.log_interval and self.processed_count % self.log_interval == 0:
            logger.info("%s/%s", self.processed_count, self.blocks_count)
        block = self.web3.eth.getBlock(self.current_block)
        self.current_block += 1
        return Block(block)



from setuptools import setup, find_packages


setup(
    name="ethereum-tools",
    version="0.1.1",
    packages=find_packages(exclude="tests"),
    description="High-level tools and library to interact with Ethereum",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Daniel Perez",
    author_email="daniel@perez.sh",
    url="https://github.com/danhper/ethereum-tools",
    download_url="https://github.com/danhper/ethereum-tools/archive/master.zip",
    license="MIT",
    install_requires=[
        "web3",
        "smart-open",
        "retry",
    ],
    extras_require={
        "dev": [
            "pylint",
            "black",
            "pytest",
        ]
    },
    entry_points={"console_scripts": ["eth-tools=eth_tools.cli:run"]},
)



from unittest.mock import PropertyMock, MagicMock, call

import pytest
from web3.types import BlockData

from eth_tools.block_iterator import BlockIterator


def test_initialization(web3):
    type(web3.eth).blockNumber = PropertyMock(return_value=100)
    block_iterator = BlockIterator(web3)
    assert block_iterator.start_block == 0
    assert block_iterator.end_block == 100
    assert block_iterator.current_block == 0
    assert block_iterator.blocks_count == 101
    assert block_iterator.processed_count == 0


def test_next(web3):
    first_block = MagicMock(spec=BlockData)
    second_block = MagicMock(spec=BlockData)
    web3.eth.getBlock.side_effect = [first_block, second_block]
    block_iterator = BlockIterator(web3, start_block=0, end_block=1)

    block = next(block_iterator)
    assert block.data == first_block
    block = next(block_iterator)
    assert block.data == second_block
    with pytest.raises(StopIteration):
        next(block_iterator)
    assert block_iterator.processed_count == 2

    web3.eth.getBlock.assert_has_calls([call(0), call(1)])



DEFAULT_BLOCK_FIELDS = [
    "number",
    "hash",
    "gasUsed",
    "gasLimit",
    "miner",
    "timestamp",
    "sha3Uncles",
    "difficulty",
    "totalDifficulty",
    "size",
    "extraData",
    "receiptsRoot",
    "stateRoot",
    "transactions_count",
]

ETHERSCAN_TRANSACTION_KEYS = [
    "blockNumber",
    "timeStamp",
    "hash",
    "nonce",
    "blockHash",
    "transactionIndex",
    "from",
    "to",
    "value",
    "gas",
    "gasPrice",
    "isError",
    "txreceipt_status",
    "input",
    "contractAddress",
    "cumulativeGasUsed",
    "gasUsed",
    "confirmations",
]

ETHERSCAN_INTERNAL_TRANSACTION_KEYS = [
    "blockNumber",
    "timeStamp",
    "hash",
    "from",
    "to",
    "value",
    "contractAddress",
    "input",
    "type",
    "gas",
    "gasUsed",
    "traceId",
    "isError",
    "errCode",
]

LOG_FORMAT = '%(asctime)-15s - %(levelname)s - %(message)s'
